/***
 * GASP: a gaming services platform for mobile multiplayer games.
 * Copyright (C) 2004 CNAM/INT
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: gasp-team@objectweb.org
 *
 * Author: Romain Pellerin
 */
package org.mega.gasp.client.doja;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Hashtable;


import javax.microedition.io.Connector;
import com.nttdocomo.io.HttpConnection;


/**
 * A game client logic running with GASP must implement this class, GASPClient.
 * 
 * @author PELLERIN Romain (pellerin@cnam.fr) - MEGA Project  
 */
public abstract class GASPClient {
	
	/*******************************************************************************
     * Explanation:
     * 
     * A Doja specific GASPClient has been created to reduce the number of methods 
     * (firstLogin, login, createAI, createPrivateAI, joinAI, joinAIRnd) because the
     * code size is critical in IApplis.
     * These methods must be delegated to a web service (another servlet independant
     * from GASP)
     * Thus, HTTP implementation is different between MIDP and Doja.
     *******************************************************************************/
    
    private static String GASPServletUrl = "http://localhost:8080/GASP/servlet/";
    protected org.mega.gasp.moods.CustomTypes customTypes;
    static protected HttpConnection connection = null;
	static private DataOutputStream dos = null;
	static private DataInputStream dis = null;
	static private Object monitor = new Object();
	

	
	/*******************************************************************************
     * Explanation:
     * 
     * A Doja specific GASPClient has been created to reduce the number of methods 
     * (firstLogin, login, createAI, createPrivateAI, joinAI, joinAIRnd) because the
     * code size is critical in IApplis.
     * These methods must be delegated to a web service (another servlet independant
     * from GASP)
     * Thus, HTTP implementation is different between MIDP and Doja.
     *******************************************************************************/
	
    abstract public void initCustomTypes();
	

	/**
     * Provide to get the events in user actorSession on the Platform. 
     *
     * @param aSID the ActorSession ID
     * @return a vector of events send by the platform
     */
    protected void getEvents(int aSID) {
		System.out.println("getEvents");
		synchronized (monitor) {
			try {
				dis = servletSendDataRequest("InGame", aSID, new Hashtable());
				eventsOf(dis);
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
				closeAllConnection();
			}
		}
	}
	
    
	/**
	 * (Try to) Start the game if the user is the owner of the application instance.
	 *  For manual mode only.
	 *  
	 * @param aSID
	 */
    protected void startAI(int aSID){
	    synchronized (monitor){
	    	String formData = "aSID="+aSID;
			try {
	            servletRequest("StartAI", formData);
	        } catch (Exception e) {
	            e.printStackTrace();
	        } finally { closeAllConnection();}
	    }
	}
	
	
	/**
	 * (Try to) End the game if the user is the owner of the application instance.
	 * For manual mode only.
	 * 
	 * @param aSID
	 */
    protected void endAI(int aSID){
    	 synchronized (monitor){
		    String formData = "aSID="+aSID;
			try {
	            servletRequest("EndAI", formData);
	        } catch (Exception e) {
	            e.printStackTrace();
	        } finally { closeAllConnection();}
    	 }
	}
    
    
    /**
	 * Quit the ApplicationInstance joined.
	 * 
	 * @param aSID the ActorSession ID
	 */
    protected void quitAI(int aSID){
    	 synchronized (monitor){
		    String formData = "aSID="+aSID;
			try {
	            servletRequest("QuitAI", formData);
	        } catch (Exception e) {
	            e.printStackTrace();
	        } finally { closeAllConnection();}
    	 }
	}
	
	
	 /**
     * Provide to send in game datas to the Platform. 
     *
     * @param aSID the ActorSession ID
     * @param h the hashtable of datas
     * @return a vector of events send by the platform
     */
    protected void sendData(int aSID, Hashtable h){
    	 synchronized (monitor){
	    	try {
	            dis = servletSendDataRequest("InGame", aSID, h);
	            eventsOf(dis);
			} 
			catch (Exception e) {
				e.printStackTrace();
			}
			finally { closeAllConnection();}
    	 }
    }

	
	/**
	 * Treats the events of the servlet response data input stream.
	 * 
     * @param dis the data input strem
     * @return
	 * @throws IOException
     */
    private void eventsOf(DataInputStream dis) throws IOException {
        byte nb = dis.readByte();
        byte type;
            
        while(nb>0){
            type = dis.readByte(); // get the type of the events
            
            switch (type){
	        	case 1: // JoinEvent
	                onJoinEvent(dis.readShort(),dis.readUTF());
	                break;
	        	case 2: // StartEvent
	        	    onStartEvent(dis.readShort());
	        	    break;
	        	case 3: // EndEvent
	        	    onEndEvent(dis.readShort());
	        	    break;
	        	case 4: // QuitEvent
	        	    onQuitEvent(dis.readShort());
	        	    break;
	        	case 5: 
	        	    try {
                        onDataEvent(dis.readShort(), customTypes.decodeData(dis));
                    } catch (IOException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    } catch (Exception e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
	        	    break;
	        	default:  break;
            }
            
            nb--;
        }
    }

	
	/**
	 * Quit the platform.
	 * 
	 * @param sID the Session ID of the user
	 */
    protected void quit(int sID) {
    	 synchronized (monitor){
	    	String formData = "sID="+sID;
			try {
	            servletRequest("Quit", formData);
	        } catch (Exception e) {
	            e.printStackTrace();
	        } finally { closeAllConnection();}
    	 }
	}

	
	/**
	 * Send a request to the specified servlet with the http post parameters.
	 * Return the servlet response DataInputStream.
	 * 
	 * @param servletName the name of the servlet
	 * @param formData the HTTP post parameters 
	 * @return the DataInputStream
	 * @throws IOException
	 */
    private DataInputStream servletRequest(String servletName, String formData) throws Exception{
       	connection = openPostConnection(GASPServletUrl+servletName);

		byte[] data = formData.getBytes();
		connection.setRequestProperty("Content-Length", Integer.toString(data.length));
			
		dos.write(data);
		dos.close();
		
		//Doja specific method
		connection.connect();
			
		if ((connection.getResponseCode() == HttpConnection.HTTP_OK)
				&&(!servletName.equals("Quit")||!servletName.equals("End")||!servletName.equals("Start"))){
				return connection.openDataInputStream();
		}
		return null;
	}


    /**
     * Send the data in the hashtable to the specified servlet.
     * 
     * @param servletName the name of the servlet
     * @return the DataInputStream of the servlet response
     * @throws IOException
     */
    private DataInputStream servletSendDataRequest(String servletName, int aSID, Hashtable h) throws Exception {
        connection = openPostConnection(GASPServletUrl+servletName);
		try {
		    dos.writeShort(aSID);
            customTypes.encodeData(h,dos);
        } catch (Throwable t) {
            t.getMessage();
            t.printStackTrace();
        }
		dos.close();
		
		//Specific Doja method
		connection.connect();
			
		if (connection.getResponseCode() == HttpConnection.HTTP_OK){
			return connection.openDataInputStream();
		}
		return null;
    }
	
    
    private static HttpConnection openPostConnection(String url ) throws Exception {
		  connection = (HttpConnection)Connector.open(url, Connector.READ_WRITE, true);
		  connection.setRequestMethod(HttpConnection.POST);
		  return connection;
	}
    
    
    /**
	 * Close all connection streams and the connection.
	 */
	static private void closeAllConnection(){
	    if (dis != null) { try {dis.close();} catch (Exception _e){}}				
		if (dos != null) { try {dos.close();} catch (Exception _e){}}
	    if (connection !=null) {try {connection.close();} catch (Exception e){}}
		connection = null;
	}
    
    
    /** Events management **/
    
    
    /**
     * Method called when a JoinEvent is received from the platform.
     * 
     * @param actorSessionID
     * @param username
     */
    abstract public void onJoinEvent(int actorSessionID, String username);
	
    
    /**
     * Method called when a StartEvent is received from the platform.
     * 
     * @param actorSessionID
     */
    abstract public void onStartEvent(int actorSessionID);

    
    /**
     * Method called when a EndEvent is received from the platform.
     * 
     * @param actorSessionID
     */
    abstract public void onEndEvent(int actorSessionID);
    

    /**
     * Method called when a QuitEvent is received from the platform.
     * 
     * @param actorSessionID
     */
    abstract public void onQuitEvent(int actorSessionID);
    
    
    /**
     * Method called when a DataEvent is received from the platform.
     * 
     * @param actorSessionID
     * @param datas the hashtable of datas
     */
    abstract public void onDataEvent(int actorSessionID, Hashtable datas);
}
